# SPDX-FileCopyrightText: 2020 Tomaz Canabrava <tcanabrava@kde.org>
# SPDX-License-Identifier: BSD-3-Clause
$XGETTEXT $(find . -name \*.cpp -o -name \*.h -o -name \*.qml) -o $podir/kcm_mobile_power.pot
