#include "handlebuttonevents.h"
#include <KPluginFactory>

K_PLUGIN_CLASS_WITH_JSON(PowerDevil::BundledActions::HandleButtonEvents, "powerdevilhandlebuttoneventsaction.json")

#include "powerdevilhandlebuttonevents_plugin.moc"
